#!/usr/bin/env python
#-*-coding: utf-8 -*-
# Copyright (c) 2019 Synology Inc. All rights reserved.

import os, sys, traceback, socket, ssl, json, hashlib, winrm, stat
import ntpath, threading, time, shutil, subprocess, struct, uuid, re
import errno
import paramiko
import requests
import codecs
from defines import *
from paramiko.ssh_exception import SSHException
from httplib import BadStatusLine
from base64 import b64encode
from scp import SCPClient, SCPException
from func_timeout import func_timeout, FunctionTimedOut

from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

def toUnixPath(path):
    if path[0] == '/':
        return path
    return path.replace('\\', '/')

def toWinPath(path):
    return path.replace('/', '\\')

def toUnixText(text):
    return text.replace('\r\n', '\n')

def escapePath(path):
    path = path.replace('`', '``')
    path = path.replace('$', '`$')
    return path

def toList(var):
    if var == None:
        return []
    if isinstance(var, list) == False:
        return [var]
    return var

class EntityNotFound(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self):
        return self.msg

class TimeoutException(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self):
        return self.msg

class AuthenticationFail(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self):
        return self.msg

class InvalidArgument(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self):
        return self.msg

class SambaMountFail(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self):
        return self.msg

class GuestOSException(Exception):
    def __init__(self, address, errno, msg):
        self.address = address
        self.errno = errno
        self.msg = msg
    def __str__(self):
        return self.msg

class ErrCertVerified(Exception):
    def __init__(self, msg, cert):
        self.msg = msg
        self.cert = cert
    def __str__(self):
        return self.msg

class PowerShellException(Exception):
    def __init__(self, exce):
        self.msg = ''
        self.category = ''
        try:
            self.exce_json = json.loads(exce)
            if 'Message' in self.exce_json['Exception']:
                self.msg = toUnixText(self.exce_json['Exception']['Message'])
            if 'ErrorIdentifier' in self.exce_json['Exception']:
                self.category = self.exce_json['Exception']['ErrorIdentifier']
            elif 'FullyQualifiedErrorId' in self.exce_json:
                self.category = self.exce_json['FullyQualifiedErrorId']
        except Exception:
            pass
        #https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.errorcategory?view=powershellsdk-1.1.0
    def __str__(self):
        return self.msg

RUN_JOB_SCRIPT \
    = '$_job = [WMI]$job.Job\n' \
    + 'if ($job.ReturnValue -eq 4096) {\n' \
    +     'while ($_job.JobState -eq 4) {\n' \
    +         'Start-Sleep -seconds 1\n' \
    +         '$_job.PSBase.Get()\n' \
    +     '}\n' \
    +     'if ($_job.JobState -ne 7) {\n' \
    +         'if ($_job.ErrorDescription -ne "") {\n' \
    +             'Throw $_job.ErrorDescription\n' \
    +         '} else {\n' \
    +             'Throw "Error Code: $_job.ErrorCode, Job State: $_job.JobState"\n' \
    +         '}\n' \
    +     '}\n' \
    + '} elseif ($job.ReturnValue -ne 0) {\n' \
    +     'Throw "Error Code: $job.ReturnValue"\n' \
    + '}\n'

class Main:
    def __init__(self, hostname, username, password, thumbprint, port, protocol, taskPool):
        self.auth_info = {
            'hostname'  : hostname,
            'username'  : username,
            'password'  : password,
            'thumbprint': thumbprint,
            'protocol'  : protocol,
            'port'      : port
        }
        self.protocol = None
        self.shell_id = None
        self.always_trust_cert = False
        self.error = None
        self.lock = threading.Lock()
        self.lockNum = 0
        self.lockNumLock = threading.Lock()

        self.task_pool = taskPool

    def __del__(self):
        self.protocol.close_shell(self.shell_id)
        print('disconnect %s' % self.auth_info['hostname'])

    def dispatch(self, command, argument):
        command = getattr(self, command)
        argument = argument or {}
        response = {}

        always_trust_cert = False
        if 'always_trust_cert' in argument:
                always_trust_cert = argument['always_trust_cert']
                argument.pop('always_trust_cert')

        try:
            func_timeout(300, self.connect, args=(always_trust_cert,))
            response['data'] = func_timeout(300, command, kwargs=argument)
            response['success'] = True
        except BaseException as e:
            traceback.print_exc(file=sys.stderr)
            response = self._parse_exception(e)
            response['success'] = False

        return response

    def _parse_exception(self, e):
        error_class = type(e).__name__
        error_category = ''
        error_message = unicode(e)
        error_params = {}

        if type(e) == PowerShellException:
            error_category = e.category
        elif type(e) == GuestOSException:
            error_params['address'] = e.address
            if e.errno > 0:
                error_message = os.strerror(e.errno)
        elif type(e) == socket.error:
            if e[0] in (errno.EHOSTUNREACH, errno.ECONNREFUSED, errno.ETIMEDOUT, errno.EHOSTDOWN, errno.EPIPE):
                error_class = 'ConnectionFail'
            error_message = e[1]
        elif type(e) in (BadStatusLine, ssl.SSLError, socket.timeout, requests.exceptions.SSLError, requests.ConnectionError, requests.Timeout):
            error_class = 'ConnectionFail'
        elif type(e) == FunctionTimedOut:
            error_message = 'WinRM operation timed out after 300 seconds.'
        elif type(e) == ErrCertVerified:
            error_class = 'ErrCertVerified'
            error_message = e.msg + ';' + e.cert
        elif isinstance(e, OSError) or type(e) == socket.gaierror:
            error_message = str(e.strerror)

        return {
            'error_class': error_class,
            'error_category': error_category,
            'error_message': error_message,
            'error_params': error_params
        }

    def _connect(self, always_trust_cert):
        if not self.auth_info['password']:
            raise AuthenticationFail('Empty password')

        cert_validation = 'validate'
        ca_path = '/etc/ssl/certs'
        if True == always_trust_cert:
            cert_validation = 'ignore'
            ca_path=''

        on_disk_thumbprint = self.auth_info['thumbprint']

        self.always_trust_cert = always_trust_cert
        self.protocol = winrm.Protocol(
            endpoint='%s://%s:%d/wsman' % (self.auth_info['protocol'], self.auth_info['hostname'], self.auth_info['port']),
            transport='ntlm',
            username=self.auth_info['username'],
            password=self.auth_info['password'],
            ca_trust_path=ca_path,
            server_cert_validation=cert_validation)

        try:
             self.shell_id = self.protocol.open_shell(codepage=65001) # utf-8
        except requests.exceptions.SSLError as e:

            cert_pem = ssl.get_server_certificate((self.auth_info['hostname'], self.auth_info['port']))
            currentThumbprint = self.getServerCert()['thumbprint']
            if currentThumbprint != on_disk_thumbprint:
                reason = 'verify_failed'
                if "hostname" in str(e) and "doesn't match" in str(e):
                    reason = 'hostname_not_match'
                print('certificate verified error!! except: %s, check_hostname: %s, on_disk_thumbprint: %s, current thumbprint: %s, always_trust_cert: %d, cert_validation: %s' % (str(e), self.auth_info['hostname'], on_disk_thumbprint, currentThumbprint, always_trust_cert, cert_validation))
                raise ErrCertVerified(msg=reason, cert = cert_pem)

            self.protocol = winrm.Protocol(
                endpoint='%s://%s:%d/wsman' % (self.auth_info['protocol'], self.auth_info['hostname'], self.auth_info['port']),
                transport='ntlm',
                username=self.auth_info['username'],
                password=self.auth_info['password'],
                server_cert_validation='ignore')
            self.shell_id = self.protocol.open_shell(codepage=65001) # utf-8

        try:
            self._run_cmd('winrm set winrm/config/service \'@{MaxConcurrentOperationsPerUser="4294967295"}\'') # workaround
        except:
            traceback.print_exc(file=sys.stderr)
        print('connect %s' % self.auth_info['hostname'])

    def connect(self, always_trust_cert):
        try:
            self.lockNumLock.acquire()
            self.lockNum += 1
            self.lockNumLock.release()
            self.lock.acquire()

            if self.error:
                raise self.error
            if self.protocol == None:
                self._connect(always_trust_cert)
        except Exception as e:
            self.error = e
            raise
        finally:
            self.lockNumLock.acquire()
            self.lockNum -= 1
            if self.lockNum == 0:
                self.error = None
            self.lockNumLock.release()
            self.lock.release()

    def getServerInfo(self, reconnect=False):
        if reconnect:
            shell_id = self.protocol.open_shell()
            self.protocol.close_shell(shell_id)

        data = {}
        data['api_type'] = 'HyperV'
        data['host_version'] = self._run_cmd('(Get-WmiObject Win32_OperatingSystem).Caption', False)
        data['host_build_ver'] = self._run_cmd('(Get-WmiObject Win32_OperatingSystem).BuildNumber', False)
        data['instance_uuid'] = self._run_cmd('Get-Wmiobject Win32_ComputerSystemProduct | Select-Object -ExpandProperty UUID', False)
        data['licensed'] = True
        data['shell_available'] = True
        data['guest_available'] = True
        return data

    def getServerCert(self):
        if self.auth_info['protocol'] != 'https':
            return {'thumbprint': ''}

        context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock = context.wrap_socket(sock, server_hostname=self.auth_info['hostname'])
        sock.connect((self.auth_info['hostname'], self.auth_info['port']))
        cert = sock.getpeercert(True)
        sock.close()

        sha1 = hashlib.sha1()
        sha1.update(cert)
        digest = sha1.hexdigest().lower()

        s, i = '', 0
        for c in digest:
            s += c
            i += 1
            if i % 2 == 0:
                s += ':'

        return {'thumbprint': s[:-1]}

    def getHostDefaultVersion(self):
        return self._run_cmd('Get-VMHostSupportedVersion -Default | Select-Object -ExpandProperty Version')

    def getHostSupportVersion(self):
        return self._run_cmd('Get-VMHostSupportedVersion | Select-Object -ExpandProperty Version')

    def getVmInfo(self, vm_id, get_detail=True):
        return self._parse_vm_data(vm_id, get_detail)

    def getDatastoreInfo(self, datastore_infos):
        return [self._parse_datastore_data(datastore['datastore_id']) for datastore in datastore_infos]

    def getEntityPath(self, entityId=None, viewType=None, attr=None):
        if viewType == 'vim.Folder' or \
            viewType == 'vim.Datastore':
            if entityId == '':
                return ''
            if not self._run_cmd('Test-Path -LiteralPath "%s" -PathType Container' % escapePath(entityId)):
                raise EntityNotFound('Cannot find folder [%s]' % entityId)

            path = entityId.split('\\')
            for i in range(1, len(path)):
                path[i] = path[i-1] + '\\' + path[i]
            return '/' + '/'.join(path)

        if viewType == 'vim.Network':
            if entityId == '':
                return ''
            self._run_cmd('Get-VMSwitch | Where {$_.Id.Guid -eq "%s"} | Select Name' % entityId)
            return '/' + entityId

        if viewType == 'vim.VirtualMachine-Host' or \
            viewType == 'vim.VirtualMachine-Folder' or \
            viewType == 'vim.VirtualMachine':
            self._run_cmd('Get-VM -Id %s | Select Name' % entityId)
            return '/' + entityId

        if viewType == 'vim.HostSystem' or \
            viewType == 'vim.ResourcePool':
            return ''

        raise InvalidArgument('invalid view type [%s]' % viewType)

    def listChildEntity(self, parentId='', viewType='', getDetail=False):
        def _collect_vms():
            return self._parse_vm_data(None, getDetail)

        def _collect_drives(drives):
            entities = []
            for drive in toList(drives):
                entities.append({
                    'id': drive['DeviceID'].lower(),
                    'name': drive['DeviceID'].lower(),
                    'type': viewType
                })
            return entities

        def _collect_folders(folders):
            entities = []
            for folder in toList(folders):
                entities.append({
                    'id': folder['FullName'].lower(),
                    'name': folder['BaseName'].lower(),
                    'type': viewType
                })
            return entities

        def _collect_networks(networks):
            entities = []
            for network in toList(networks):
                entities.append({
                    'id': network['Id'],
                    'name': network['Name'],
                    'type': 'vim.Network'
                })
            return entities

        def _collect_hosts():
            hostName = self._run_cmd('$env:computername', False)
            return [{
                'id': hostName,
                'name': hostName,
                'type': 'vim.HostSystem',
                'migratable': self.getMigrateCapability()
            }]

        if parentId == '':
            if viewType == 'vim.Folder' or \
               viewType == 'vim.Datastore':
                drives = self._run_cmd('Get-WMIObject -Query "Select DeviceID From win32_logicaldisk Where DriveType=3"')
                entities = _collect_drives(drives)
            elif viewType == 'vim.Network':
                networks = self._run_cmd('Get-VMSwitch | Select Id,Name')
                entities = _collect_networks(networks)
            elif viewType == 'vim.VirtualMachine' or \
                viewType == 'vim.VirtualMachine-Host' or \
                viewType == 'vim.VirtualMachine-Folder':
                entities = _collect_vms()
            elif viewType == 'vim.ResourcePool':
                entities = _collect_hosts()
            else:
                raise InvalidArgument('invalid view type [%s]' % viewType)
        else:
            if viewType == 'vim.Folder' or \
               viewType == 'vim.Datastore':
                folders = self._run_cmd('Get-ChildItem -Force -Directory -LiteralPath "%s/" | Select FullName,BaseName' % escapePath(parentId))
                entities = _collect_folders(folders)
            else:
                raise InvalidArgument('invalid view type [%s]' % viewType)

        entities.sort(key=lambda obj: (obj['type'], obj['name']))
        return entities

    def registerVm(self,
            vm_name=None,
            folder_id=None,
            resourcepool_id=None,
            host_id=None,
            os_name=None,
            cpu_number=None,
            memory_size=None,
            is_system_mbr=None,
            vdisks=None,
            networks=None,
            file_path=None,
            gen_new_uuid=False):

        cmd = '$vdiskPaths = @("' + '","'.join([escapePath(vdisk['filename']) for vdisk in vdisks]) + '")\n' \
            + '$vdiskUuids = @("' + '","'.join([vdisk['lunUuids'] for vdisk in vdisks]) + '")\n' \
            + '$vdiskBoots = @(' + ','.join([str(int(vdisk['bootable'])) for vdisk in vdisks]) + ')\n'

        if 0 < len(networks):
            cmd += '$networks = @("' + '","'.join([network['name'] if network['name'] else '' for network in networks]) + '")\n'
        else:
            cmd += '$networks = @()\n'

        cmd += '$movedItems = @()\n' \
            + '$createdItems = @()\n' \
            + ('$vmName = "%s"\n' % vm_name) \
            + ('$cfgPath = "%s"\n' % escapePath(folder_id)) \
            + ('$vmxPath = "%s"\n' % escapePath(file_path)) \
            + 'for ($i=0; $i -lt $vdiskPaths.length; ++$i) {\n' \
            +     'if ($vdiskUuids[$i] -ne "") {\n' \
            +         'continue\n' \
            +     '}\n' \
            +     '$vdiskItem = Get-Item -LiteralPath $vdiskPaths[$i]\n' \
            +     '$diskHasData = Join-Path $vdiskItem.Directory ("abb_disk_has_data_" + $i)\n' \
            +     'if (-Not (Test-Path -LiteralPath $diskHasData)) {\n' \
            +         'New-Item $diskHasData | Out-Null\n' \
            +         '$createdItems += ,$diskHasData\n' \
            +         'continue\n' \
            +     '}\n' \
            +     '$newFolder = Split-Path $vdiskItem.Directory -Parent\n' \
            +     '$newBaseName = $vdiskItem.Basename\n' \
            +     'for ($num = 1; Test-Path -LiteralPath (Join-Path $newFolder ($newBaseName + $vdiskItem.Extension)); ++$num) {\n' \
            +         '$newBaseName = $vdiskItem.Basename + "_$num"\n' \
            +     '}\n' \
            +     '$vdiskPaths[$i] = Join-Path $newFolder ($newBaseName + $vdiskItem.Extension)\n' \
            +     'Move-Item -LiteralPath $vdiskItem.FullName $vdiskPaths[$i]\n' \
            +     '$movedItems += ,@($vdiskItem.FullName, $vdiskPaths[$i])\n' \
            + '}\n'

        vmxExisted = self._run_cmd('Test-Path -LiteralPath "%s"' % escapePath(file_path))
        if vmxExisted:
            genMacCmd \
                = 'if ($config.VM.NetworkAdapters[$i].DynamicMacAddressEnabled -eq $false) {\n' \
                +     '$newMacAddress = "00" + ((0..9 | %{"{0:x}" -f (Get-Random -Minimum 0 -Maximum 15)}) -join "")\n' \
                +     'Set-VMNetworkAdapter -VMNetworkAdapter $config.VM.NetworkAdapters[$i] -StaticMacAddress $newMacAddress\n' \
                + '}\n'
            addDiskCmd = 'Add-VMHardDiskDrive -VM $vm -ControllerType $hardDrives[$i].ControllerType -ControllerNumber $hardDrives[$i].ControllerNumber -ControllerLocation $hardDrives[$i].ControllerLocation'

            cmd += '$config = Compare-VM -Copy -GenerateNewId -Path $vmxPath -VirtualMachinePath $cfgPath -SnapshotFilePath $cfgPath -SmartPagingFilePath $cfgPath\n' \
                + '$hardDrives = $config.VM.HardDrives\n' \
                + 'Rename-VM -VM $config.VM -NewName $vmName\n' \
                + 'try { Set-VM -VM $config.VM -AutomaticCheckpointsEnabled $false } catch {}\n' \
                + 'Get-VMSnapshot $config.VM | %{ try { Remove-VMSnapshot $_ } catch {} }\n' \
                + 'Remove-VMHardDiskDrive -VMHardDiskDrive $hardDrives\n' \
                + 'for ($i=0; $i -lt $networks.length; ++$i) {\n' \
                +     'if (!$networks[$i]) {\n' \
                +         'continue\n' \
                +     '}\n' \
                +     'Connect-VMNetworkAdapter -VMNetworkAdapter $config.VM.NetworkAdapters[$i] -SwitchName $networks[$i]\n' \
                +      (genMacCmd if gen_new_uuid else '') \
                + '}\n' \
                + '$vm = Import-VM -CompatibilityReport $config\n' \
                + 'for ($i=0; $i -lt $hardDrives.length; ++$i) {\n' \
                +     'if ($vdiskUuids[$i] -ne "") {\n' \
                +         '$disk = Get-WmiObject Win32_DiskDrive | Where SerialNumber -eq $vdiskUuids[$i] | Select -ExpandProperty Index\n' \
                +         '$cmd = @"\n' \
                +         'select disk $disk\n' \
                +         'offline disk\n' \
                +         '"@\n' \
                +         '$cmd | diskpart | Out-Null\n' \
                +          addDiskCmd + ' -DiskNumber $disk\n' \
                +     '} else {\n' \
                +          addDiskCmd + ' -Path $vdiskPaths[$i]\n' \
                +     '}\n' \
                + '}\n' \
                + 'if ($vm.Generation -eq 1) {\n' \
                +     'Set-VMBios -VM $vm -StartupOrder @("IDE", "CD", "LegacyNetworkAdapter", "Floppy")\n' \
                + '} else {\n' \
                +     'Get-VMHardDiskDrive -VM $vm | %{\n' \
                +         'Set-VMFirmware -VM $vm -FirstBootDevice $_\n' \
                +     '}\n' \
                + '}\n'
        else:
            addDiskCmd = 'Add-VMHardDiskDrive -VMDriveController $ctrl -ControllerLocation $unit'

            cmd += '$generation = ' + ('1' if is_system_mbr else '2') + '\n' \
                + '$vm = New-VM -Name $vmName -MemoryStartupBytes ' + str(memory_size) + 'MB -Generation $generation\n' \
                + 'Set-VMProcessor -VM $vm -Count ' + str(cpu_number) + '\n' \
                + 'try { Set-VM -VM $vm -AutomaticCheckpointsEnabled $false } catch {}\n' \
                + ('' if 'windows' in os_name.lower() else 'try { Set-VMFirmware -VM $vm -EnableSecureBoot On -SecureBootTemplate MicrosoftUEFICertificateAuthority } catch {}\n') \
                + 'Remove-VMNetworkAdapter -VM $vm\n' \
                + 'for ($i=0; $i -lt $vdiskPaths.length; ++$i) {\n' \
                +     'if ($vdiskBoots[$i] -and $generation -eq 1) {\n' \
                +         '$ctrl = (Get-VMIdeController -VM $vm)[0]\n' \
                +         '$unit = 0\n' \
                +     '} else {\n' \
                +         '$ctrl = Get-VMScsiController -VM $vm\n' \
                +         '$unit = $i\n' \
                +     '}\n' \
                +     'if ($vdiskUuids[$i] -ne "") {\n' \
                +         '$disk = Get-WmiObject Win32_DiskDrive | Where SerialNumber -eq $vdiskUuids[$i] | Select -ExpandProperty Index\n' \
                +         '$cmd = @"\n' \
                +         'select disk $disk\n' \
                +         'offline disk\n' \
                +         '"@\n' \
                +         '$cmd | diskpart | Out-Null\n' \
                +          addDiskCmd + ' -DiskNumber $disk\n' \
                +     '} else {\n' \
                +          addDiskCmd + ' -Path $vdiskPaths[$i]\n' \
                +     '}\n' \
                + '}\n' \
                + 'for ($i=0; $i -lt $networks.length; ++$i) {\n' \
                +     'Add-VMNetworkAdapter -VM $vm -SwitchName $networks[$i]\n' \
                + '}\n' \
                + '$specPath = $vmxPath -replace (".vmcx$", "")\n' \
                + '$specItem = Get-ChildItem -LiteralPath $vm.Path -Filter ($vm.Id.Guid + ".*") -File -Recurse\n' \
                + '$specItem | %{ Copy-Item -LiteralPath $_.FullName ($specPath + $_.Extension) }\n'

        cmd = 'try {\n' \
            +     cmd + '$vm.Id.Guid\n' \
            + '} catch {\n' \
            +     'try { Remove-VM -VM $vm -Force } catch {}\n' \
            +     'try { Remove-Item -LiteralPath $createdItems } catch {}\n' \
            +     'try { $movedItems | %{ Move-Item -LiteralPath $_[1] $_[0] }} catch {}\n' \
            +     'throw\n' \
            + '}'

        return self._run_cmd(cmd, False, True, True)

    def uploadFile(self, datacenter_id=None, datastore_id=None, src_path=None, dst_path=None):
        options = 'noserverino,iocharset=utf8,password=%s,uid=%d,gid=%d,rw' % \
            (self.auth_info['password'],
            int(os.environ['PKG_UID']),
            int(os.environ['PKG_GID']))

        username = self.auth_info['username'].split('\\')
        if len(username) == 2:
            options += ',domain=%s,username=%s' % (username[0], username[1])
        else:
            options += ',domain=WORKGROUP,username=%s' % (username[0])

        src_path = src_path.encode('utf-8')
        dst_path = os.path.join(datastore_id, dst_path).encode('utf-8')
        host_dir = ('//%s/%s' % (self.auth_info['hostname'], ntpath.dirname(dst_path.replace(':', '$', 1)))).encode('utf-8')
        temp_dir = os.path.join(VM_SYNOABK_TEMP, randomString()).encode('utf-8')
        temp_path = os.path.join(temp_dir, ntpath.basename(dst_path)).encode('utf-8')

        if 0 <= dst_path.find(','):
            raise InvalidArgument('Invalid character "," in path [%s]' % dst_path)

        try:
            os.mkdir(temp_dir)

            ret = subprocess.Popen(['/bin/mount.cifs', '-o', options+',vers=syno', host_dir, temp_dir], stderr=subprocess.PIPE).communicate()
            if ret[1]:
                ret = subprocess.Popen(['/bin/mount.cifs', '-o', options, host_dir, temp_dir], stderr=subprocess.PIPE).communicate()
                if ret[1]:
                    raise SambaMountFail(ret[1])

            shutil.copyfile(src_path, temp_path)
        finally:
            try:
                subprocess.Popen(['/bin/umount', '-l', temp_dir], stderr=subprocess.PIPE).communicate()
            except:
                pass
            try:
                os.rmdir(temp_dir)
            except:
                pass

    def createDatastore(self, name=None, host_id=None, ip_address=None, mount_path=None):
        dst_root = self._run_cmd('echo $env:LOCALAPPDATA/ActiveBackup', False)
        dst_path = ntpath.join(dst_root, name + '-' + randomString())
        self.createFolder(folder_path=dst_path)
        return self._parse_datastore_data(dst_path)

    def removeDatastore(self, host_id=None, datastore_id=None):
        self.removeFolder(folder_path=datastore_id)

    def createFolder(self, datacenter_id=None, folder_path=None):
        self._run_cmd('New-Item -ItemType Directory -Path "%s"' % escapePath(folder_path), False)

    def removeFolder(self, datacenter_id=None, folder_path=None):
        self._run_cmd('Remove-Item -LiteralPath "%s" -Recurse -Force' % escapePath(folder_path), False)

    def renameFolder(self, datacenter_id=None, src_path=None, dst_path=None):
        self._run_cmd('Rename-Item -LiteralPath "%s" -NewName "%s"' % (escapePath(src_path), escapePath(dst_path)), False)

    def createDisk(self, file_path=None, thin_provision=None, capacity=None):
        disk_type = 'dynamic' if thin_provision else 'fixed'
        if capacity < 3145728:
            capacity = 3145728
        return self.runTask('New-VHD -%s -SizeBytes %s -Path "%s"' % (disk_type, capacity, escapePath(file_path)))

    def testConnection(self, ip=None, port=None):
        cmd = 'New-Object System.Net.Sockets.TCPClient -ArgumentList %s, %d | Select Connected'
        try:
            return self._run_cmd(cmd % (ip, port))["Connected"]
        except:
            return False

    def isIscsiRunning(self):
        return self._run_cmd('(Get-Service msiscsi).Status -eq "Running"')

    def checkHypervisorRequirements(self):
        try:
            self._run_cmd('Get-Cluster', False)
            part_of_domain = True
        except:
            part_of_domain = False

        samba1_enabled = self._run_cmd('Get-SmbServerConfiguration | Select -ExpandProperty EnableSMB1Protocol')
        samba2_enabled = self._run_cmd('Get-SmbServerConfiguration | Select -ExpandProperty EnableSMB2Protocol')
        script_can_run = not self._run_cmd('Try { $p="$env:temp/ActiveBackup-' + randomString() + '.ps1"; New-Item $p | Out-Null; &$p } Catch { $_ } Finally { Remove-Item $p }', False)

        all_drives = self._run_cmd('$a=""; [System.IO.DriveInfo]::GetDrives() | Where DriveType -Ne "CDROM" | %{ $a+=$_.Name[0] }; $a', False)
        net_drives = self._run_cmd('$a=""; Get-SmbShare | Where { $_.Name -Match \'^\\S\\$$\' } | %{ $a+=$_.Name[0] }; $a', False)
        invalid_shares = list(set(all_drives.upper()) - set(net_drives.upper()) - set('AB'))

        samba_mountable = True
        src_path = self._run_cmd('echo $env:LOCALAPPDATA', False).replace(':', '$', 1)
        host_dir = ('//%s/%s' % (self.auth_info['hostname'], src_path)).encode('utf-8')
        temp_dir = os.path.join(VM_SYNOABK_TEMP, randomString()).encode('utf-8')
        options = 'noserverino,iocharset=utf8,password=%s,uid=%d,gid=%d,ro' % \
            (self.auth_info['password'],
            int(os.environ['PKG_UID']),
            int(os.environ['PKG_GID']))

        username = self.auth_info['username'].split('\\')
        if len(username) == 2:
            options += ',domain=%s,username=%s' % (username[0], username[1])
        else:
            options += ',domain=WORKGROUP,username=%s' % (username[0])

        try:
            os.mkdir(temp_dir)

            ret = subprocess.Popen(['/bin/mount.cifs', '-o', options+',vers=syno', host_dir, temp_dir], stderr=subprocess.PIPE).communicate()
            if ret[1]:
                ret = subprocess.Popen(['/bin/mount.cifs', '-o', options, host_dir, temp_dir], stderr=subprocess.PIPE).communicate()
                if ret[1]:
                    samba_mountable = False
        finally:
            try:
                subprocess.Popen(['/bin/umount', '-l', temp_dir], stderr=subprocess.PIPE).communicate()
            except:
                pass
            try:
                os.rmdir(temp_dir)
            except:
                pass

        return {
            'part_of_domain': part_of_domain,
            'samba1_enabled': samba1_enabled,
            'samba2_enabled': samba2_enabled,
            'script_can_run': script_can_run,
            'invalid_shares': invalid_shares,
            'samba_mountable': samba_mountable
        }

    def createIscsiPortal(self, portal_addr=None):
        try:
            self._run_cmd('Get-IscsiTargetPortal -TargetPortalAddress ' + portal_addr + ' | Update-IscsiTargetPortal', True)
            return False
        except PowerShellException:
            self._run_cmd('New-IscsiTargetPortal -TargetPortalAddress ' + portal_addr, False)
        return True

    def updateIscsiTargetRemovePortal(self, portal_addr=None, remove_portal=False):
        cmd = 'Get-IscsiTargetPortal -TargetPortalAddress ' + portal_addr + ' | Update-IscsiTargetPortal'
        if remove_portal:
            cmd += '\nRemove-IscsiTargetPortal -TargetPortalAddress ' + portal_addr + ' -Confirm:$false'

        self.runProcess(cmd, True)

    def createIscsiDisk(self, node_addr=None):
        try:
            self._run_cmd('Connect-IscsiTarget -NodeAddress ' + node_addr, False)
        except PowerShellException as e:
            if 0 > e.category.find('0xefff003f'):
                raise e

    def removeIscsiDisk(self, node_addr=None):
        self._run_cmd('Disconnect-IscsiTarget -NodeAddress ' + node_addr + ' -Confirm:$false', True)
        #refresh hyperv iscsi cache
        self.runProcess('Update-IscsiTarget', True)

    def migrateVm(self, vm_id=None, datastore_id=None):
        pass

    def poweronVm(self, vm_id=None):
        self._run_cmd('Start-VM -VM (Get-VM -Id %s)' % vm_id)

    def poweroffVm(self, vm_id=None):
        self._run_cmd('Stop-VM -VM (Get-VM -Id %s) -TurnOff' % vm_id)

    def unregisterVm(self, vm_id=None):
        self._run_cmd('Remove-VM -VM (Get-VM -Id %s) -Force' % vm_id)

    def listSnapshot(self, vm_id=None):
        return self._parse_snapshot_data(vm_id)

    def getSnapshot(self, vm_id=None, snap_id=None):
        return self._parse_snapshot_data(vm_id, snap_id)

    def takeSnapshot(self, vm_id=None, snap_name=None, snap_desc=None, quiesce=False):
        vm = self._run_cmd('Get-VM -Id ' + vm_id + '| Select OperationalStatus,StatusDescriptions')
        if len(vm['OperationalStatus']) != 1 or vm['OperationalStatus'][0] != 2: # status != OK
            raise Exception('Virtual Machine is in busy status "%s".' % vm['StatusDescriptions'][-1])

        cmd = '$service = Get-WmiObject -Namespace root\\virtualization\\v2 -Class Msvm_VirtualSystemSnapshotService\n' \
            + '$vm = Get-WmiObject -Namespace root\\virtualization\\v2 -Class Msvm_ComputerSystem -Filter "Name=\'' + vm_id + '\'"\n' \
            + '$setting = ([WMIClass]"\\\\.\\root\\virtualization\\v2:Msvm_VirtualSystemSnapshotSettingData").CreateInstance()\n' \
            + '$setting.ConsistencyLevel = ' + ('2' if quiesce else '0') + '\n' \
            + '$setting.IgnoreNonSnapshottableDisks = $true\n' \
            + '$job = $service.CreateSnapshot($vm, $setting.GetText(2), 32768)\n' \
            + RUN_JOB_SCRIPT \
            + '$snapshot = ($_job.GetRelated("Msvm_VirtualSystemSettingData") | %{$_})\n' \
            + '$snapshotId = $snapshot.InstanceId.Split(":")[1]\n' \
            + 'Rename-VMSnapshot -VMSnapshot (Get-VMSnapshot -Id $snapshotId) -NewName "' + snap_name + '"\n' \
            + '$snapshotId'

        try:
            snap_id = self._run_cmd(cmd, False, True, True)
            return self._parse_snapshot_data(vm_id, snap_id)
        except Exception as e:
            try:
                self.removeSnapshot(vm_id, snap_id)
            except:
                pass
            raise e

    def removeSnapshot(self, vm_id=None, snap_id=None, nonblocking=None):
        if nonblocking:
            self.runProcess('Remove-VMSnapshot -VMSnapshot (Get-VMSnapshot -Id %s)' % snap_id, True)
            return None

        snapshot = self._parse_snapshot_data(vm_id, snap_id)

        if snapshot['type'] != 1: # recovery
            self._run_cmd('Remove-VMSnapshot -VMSnapshot (Get-VMSnapshot -Id %s)' % snap_id, False)
        else:
            cmd = '$key2Uuid = @{}\n' \
                + '$uuid2cbt = @{}\n' \
                + 'Get-VM -Id ' + vm_id + ' | Get-VMHardDiskDrive | %{\n' \
                +     'Try {\n' \
                +         '$path = ($_|Get-VHD).Path\n' \
                +         'while ($path) {\n' \
                +             '$disk = Get-VHD -Path $path\n' \
                +             '$path = $disk.ParentPath\n' \
                +         '}\n' \
                +         '$key2Uuid[($_.Id -Replace ".$")] = $disk.DiskIdentifier\n' \
                +     '} Catch {}\n' \
                + '}\n' \
                + '$snapshot = Get-WmiObject -Namespace root\\virtualization\\v2 -Class Msvm_VirtualSystemSettingData -Filter "InstanceID=\'Microsoft:' + snap_id + '\'"\n' \
                + '$service = Get-WmiObject -Namespace root\\virtualization\\v2 -Class Msvm_VirtualSystemSnapshotService\n' \
                + '$job = $service.ConvertToReferencePoint($snapshot)\n' \
                + RUN_JOB_SCRIPT \
                + '$refpoint = ($_job.GetRelated("Msvm_VirtualSystemReferencePoint") | %{$_})\n' \
                + 'for ($i = 0; $i -lt $refpoint.VirtualDiskIdentifiers.Count; $i++) {\n' \
                +     'Try {\n' \
                +         '$key = $refpoint.VirtualDiskIdentifiers[$i] -Replace ".$"\n' \
                +         '$cbt = $refpoint.ResilientChangeTrackingIdentifiers[$i]\n' \
                +         '$uuid2cbt[$key2Uuid[$key]] = $cbt\n' \
                +     '} Catch {}\n' \
                + '}\n' \
                + '$uuid2cbt'

            uuid2cbt = self._run_cmd(cmd, True, True, True)
            for vdisk in snapshot['virtual_disk']:
                try:
                    vdisk['change_id'] = uuid2cbt[vdisk['key']]
                except:
                    pass
        return snapshot

    def runTask(self, cmd=None):
        cmdJob = '$job = ' + cmd + ' -asJob\n' \
            + '$file = Join-Path $env:TEMP ("ActiveBackupTask_" + $PID)\n' \
            + 'While (($job.JobStateInfo.state -eq 0) -or ($job.JobStateInfo.state -eq 1)) {\n' \
            + '     $workaround = $job | Select *\n' \
            + '     $job.progress | ConvertTo-Json > $file\n' \
            + '     $job | Wait-Job -Timeout 1\n' \
            + '}\n' \
            + '$job.Error | ConvertTo-Json > $file'

        return self.runProcess(cmdJob, True)

    def runProcess(self, cmd=None, is_ps_script=False):
        if is_ps_script:
            cmd = b64encode(cmd.encode('utf_16_le')).decode('ascii')
            cmd = 'powershell -encodedcommand ' + cmd
        cmd = "Invoke-WmiMethod -Path win32_process -Name create -ArgumentList '%s'" % cmd
        return self._run_cmd(cmd)['ProcessId']

    def getTaskInfo(self, task_id=None):
        if task_id in self.task_pool:
            task = self._get_task(task_id)

            if task.info.state == 'error':
                self._del_task(task_id)
                raise task.info.error

            if task.info.state == 'success':
                self._del_task(task_id)

            return {
                'success': task.info.state == 'success',
                'progress': task.info.progress,
                'result': task.info.result
            }
        else:
            cmd = '$file = Join-Path $env:TEMP ActiveBackupTask_' + task_id + '\n'\
                + '$result = @{}\n' \
                + '$proc = Get-Process -Id ' + task_id + ' -ErrorAction SilentlyContinue\n' \
                + 'if ($proc) {\n' \
                + '     $result["success"] = $FALSE;\n' \
                + '     if (Test-Path $file) {\n' \
                + '         $progress = Get-Content $file | Out-String | ConvertFrom-Json\n' \
                + '         $result["progress"] = $progress.PercentComplete;\n' \
                + '     } else {\n' \
                + '         $result["progress"] = 0;\n' \
                + '     }\n' \
                + ' } else {\n' \
                + '     $result["success"] = $TRUE;\n' \
                + '     $result["progress"] = 100;\n' \
                + '     if (Test-Path $file) {\n' \
                + '         $progress = Get-Content $file | Out-String | ConvertFrom-Json\n' \
                + '         Remove-Item $file\n' \
                + '         if ($progress.Exception) {\n' \
                + '             $result["exception"] = $progress\n' \
                + '         }\n' \
                + '     }\n' \
                + ' }\n' \
                + '$result'

            ret = self._run_cmd(cmd, True, True)
            if 'exception' in ret:
                raise PowerShellException(json.dumps(ret['exception']))
            return ret

    def cancelTask(self, task_id=None):
        if task_id in self.task_pool:
            task = self._get_task(task_id)
            task.CancelTask()
            self._del_task(task_id)
        else:
            self._run_cmd('Stop-Process -Id %s' % task_id)

    def getCbtState(self, vm_id=None):
        return {
            'support': True,
            'enabled': True
        }

    def setCbtState(self, vm_id=None, enable=False):
        pass

    def uploadFileInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, src_path=None, dst_path=None, overwrite=True):
        def uploadFileInVmTask(task):
            auth_info = {
                'vm_id': vm_id,
                'address': vm_address,
                'port': vm_port,
                'username': vm_user,
                'password': vm_pwd
            }

            def func(transport, src_path, dst_path):
                sftp = paramiko.SFTPClient.from_transport(transport)
                # put file hang workaround
                sftp.ultra_debug = True
                src_path = src_path.encode('utf-8')
                dst_path = toUnixPath(dst_path.encode('utf-8'))
                temp_dst_path = dst_path + '_abb.' + randomString()

                def updateProgress(filename, total, transferred):
                    if task._canceled:
                        try:
                            sftp.remove(temp_dst_path)
                        except Exception:
                            print('remove temp. file [{}] failed.'.format(temp_dst_path))
                        raise Exception('cancel')
                    task.info.progress = 100.0 * transferred / total

                def uploadSFTP():
                    dst_stat = None

                    try:
                        dst_stat = sftp.stat(dst_path).st_mode
                    except IOError as e:
                        if e.errno != errno.ENOENT:
                            task.info.state = 'error'
                            task.info.error = e
                            return
                    else:
                        if not overwrite:
                            task.info.progress = 100.0
                            task.info.state = 'success'
                            return
                        elif not stat.S_ISREG(dst_stat):
                            raise GuestOSException(vm_address, -1, 'Failed to overwrite file/folder. Please make sure the file/folder property is the same.')

                    try:
                        scp = SCPClient(transport, progress=updateProgress)
                        try:
                            scp.put(src_path, temp_dst_path)
                        except IOError as e:
                            raise GuestOSException(vm_address, -1, 'Cannot access the specified device, path, or file.')
                        except SCPException as e:
                            if task.info.progress == 100.0:
                                raise GuestOSException(vm_address, -1, 'There is not enough space to upload the file. [%s]' % dst_path)
                            raise
                        finally:
                            scp.close()

                        sftp.chmod(temp_dst_path, stat.S_IRWXU)

                        if overwrite:
                            try:
                                sftp.remove(dst_path)
                            except IOError as e:
                                if e.errno != errno.ENOENT:
                                    raise
                        sftp.rename(temp_dst_path, dst_path)
                        task.info.state = 'success'
                    except Exception as e:
                        task.info.state = 'error'
                        task.info.error = e
                        try:
                            sftp.remove(temp_dst_path)
                        except Exception:
                            print('remove temp. file [{}] failed.'.format(temp_dst_path))
                        raise e

                def uploadWindows():
                    device_id = ntpath.splitdrive(dst_path)[0]
                    temp_dst_folder = os.path.join(device_id, 'Synology')
                    temp_dst_path = os.path.join(temp_dst_folder, randomString() + '.syno_abb_tmp')
                    dst_dirname = ntpath.dirname(dst_path)
                    dst_basename = ntpath.basename(dst_path)

                    try:
                        list_result = self.listFilesInVm(vm_id, vm_user, vm_pwd, vm_address, vm_port, dst_dirname)
                    except IOError as e:
                        if e.errno != errno.ENOENT:
                            task.info.state = 'error'
                            task.info.error = e
                            return

                    for f in list_result['files']:
                        if f['name'] == dst_basename:
                            if not overwrite:
                                task.info.progress = 100.0
                                task.info.state = 'success'
                                return
                            elif f['is_dir']:
                                raise GuestOSException(vm_address, -1, 'Failed to overwrite file/folder. Please make sure the file/folder property is the same.')
                            break

                    try:
                        try:
                            sftp.mkdir(temp_dst_folder)
                        except Exception as e:
                            print('mkdir. folder [%s] failed. err[%s]' %(temp_dst_path, str(e)))

                        scp = SCPClient(transport, progress=updateProgress)
                        try:
                            scp.put(src_path, temp_dst_path)
                        except IOError as e:
                            raise GuestOSException(vm_address, -1, 'Cannot access the specified device, path, or file.')
                        except SCPException as e:
                            if task.info.progress == 100.0:
                                raise GuestOSException(vm_address, -1, 'There is not enough space to upload the file. [%s]' % dst_path)
                            raise
                        finally:
                            scp.close()

                        sftp.chmod(temp_dst_path, stat.S_IRWXU)

                        self.runGuestShell(auth_info, 'move /y "%s" "%s"' % (toWinPath(temp_dst_path), toWinPath(dst_path)), 180 , True)
                        task.info.state = 'success'
                    except Exception as e:
                        task.info.state = 'error'
                        task.info.error = e
                        try:
                            sftp.remove(temp_dst_path)
                        except Exception:
                            print('remove temp. file [{}] failed.'.format(temp_dst_path))
                        raise e

                if self.isWindowsVM(vm_id, vm_user, vm_pwd, vm_address, vm_port):
                    uploadWindows()
                else:
                    uploadSFTP()


            return self.runTaskGuestSFTP(task, func, auth_info, src_path, dst_path)

        class FakeTaskInfo:
            def __init__(self):
                self.key = 'upload_to_vm_' + uuid.uuid4().hex[:12]
                self.state = 'running'
                self.error = None
                self.result = None
                self.progress = 0

        class FakeTask:
            def __init__(self):
                self.info = FakeTaskInfo()
                self._canceled = False
            def CancelTask(self):
                self._canceled = True

        task = FakeTask()
        threading.Thread(target=uploadFileInVmTask, args=(task,)).start()
        return self._put_task(task)

    def isWindowsVM(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None):
        auth_info = {
            'vm_id': vm_id,
            'address': vm_address,
            'port': vm_port,
            'username': vm_user,
            'password': vm_pwd
        }
        isWindows = False
        try:
            results = self.runGuestShell(auth_info, 'ver', 180, True)
            if "Microsoft Windows" in results:
                isWindows = True
        except:
            self.runGuestShell(auth_info, 'uname', 180)
            isWindows = False
        return isWindows

    def listFilesInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, file_path=None, index=None, max_results=None, match_pattern=None):
        auth_info = {
            'vm_id': vm_id,
            'address': vm_address,
            'port': vm_port,
            'username': vm_user,
            'password': vm_pwd
        }

        def func(transport, file_path):
            sftp = paramiko.SFTPClient.from_transport(transport)

            out = {'files': [], 'remaining': 0}

            if not file_path:
                if self.isWindowsVM(vm_id, vm_user, vm_pwd, vm_address, vm_port):
                    results = self.runGuestShell(auth_info, 'wmic logicaldisk where drivetype=3 get DeviceID')
                    drives = results.split()[1:]
                    for drive in toList(drives):
                        out['files'].append({
                            'name': drive.lower(),
                            'path': drive.lower(),
                            'size': 0,
                            'is_dir': True,
                            'mtime': 0
                        })
                else:
                    out['files'].append({
                        'name': '/',
                        'path': '/',
                        'size': 0,
                        'is_dir': True,
                        'mtime': 0
                    })
            else:
                if self.isWindowsVM(vm_id, vm_user, vm_pwd, vm_address, vm_port):
                    win_file_path = toWinPath(file_path) + '\\'
                    results = self.runGuestShell(auth_info, 'dir "%s"' % win_file_path, 180 , True)

                    skipDir = [".", ".."]
                    files = results.split('\r\n')[5:-2]
                    for f in files:
                        dd = f.split(None, 4)
                        filename = dd[4]
                        if filename in skipDir:
                            continue

                        isDir = False
                        size = 0
                        if "<DIR>" in dd[3]:
                            isDir = True
                        else:
                            size = dd[3].replace(",", "")

                        out['files'].append({
                            'name': filename,
                            'path': win_file_path + filename,
                            'size': size,
                            'is_dir': isDir,
                            'mtime': 0
                        })
                else:
                    results = sftp.listdir_attr(file_path)

                    for result in results:
                        out['files'].append({
                            'name': result.filename,
                            'path': file_path + '/' + result.filename,
                            'size': result.st_size,
                            'is_dir': stat.S_ISDIR(result.st_mode),
                            'mtime': result.st_mtime
                        })
            return out

        return self.runGuestSFTP(func, auth_info, file_path)

    def downloadFileFromVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, src_path=None, dst_path=None):
        auth_info = {
            'vm_id': vm_id,
            'address': vm_address,
            'port': vm_port,
            'username': vm_user,
            'password': vm_pwd
        }

        def func(transport, src_path, dst_path):
            scp = SCPClient(transport)
            src_path = src_path.encode('utf-8')
            dst_path = dst_path.encode('utf-8')
            scp.get(toUnixPath(src_path), dst_path)
            scp.close()
            os.chown(dst_path, int(os.environ['PKG_UID']), int(os.environ['PKG_GID']))
            os.chmod(dst_path, stat.S_IRWXU)

        self.runGuestSFTP(func, auth_info, src_path, dst_path)

    def removeFileInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, file_path=None):
        auth_info = {
            'vm_id': vm_id,
            'address': vm_address,
            'port': vm_port,
            'username': vm_user,
            'password': vm_pwd
        }

        def func(transport, file_path):
            if self.isWindowsVM(vm_id, vm_user, vm_pwd, vm_address, vm_port):
                self.runGuestShell(auth_info, 'del "%s"' % toWinPath(file_path), 180 , True)
            else :
                sftp = paramiko.SFTPClient.from_transport(transport)
                sftp.unlink(file_path)

        self.runGuestSFTP(func, auth_info, file_path)

    def readEnvVariableInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, names=None):
        auth_info = {
            'vm_id': vm_id,
            'address': vm_address,
            'port': vm_port,
            'username': vm_user,
            'password': vm_pwd
        }

        results = self.runGuestShell(auth_info, 'cmd.exe /c set')
        variables = []
        for result in results.split():
            try:
                key, value = result.split('=')
                if key == names:
                    variables.append(result)
            except:
                pass
        return variables

    def executeProgramInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, file_path=None, arguments=None, timeout=180):
        auth_info = {
            'vm_id': vm_id,
            'address': vm_address,
            'port': vm_port,
            'username': vm_user,
            'password': vm_pwd
        }

        out_json = {}
        try:
            self.runGuestShell(auth_info, '%s %s' % (file_path, arguments), timeout)
        except GuestOSException as e:
            out_json['exit_code'] = e.errno
            out_json['exception'] = str(e)
        else:
            out_json['exit_code'] = 0

        return out_json

    def createDirectoryInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, dir_path=None):
        auth_info = {
            'vm_id': vm_id,
            'address': vm_address,
            'port': vm_port,
            'username': vm_user,
            'password': vm_pwd
        }

        if self.isWindowsVM(vm_id, vm_user, vm_pwd, vm_address, vm_port):
            dir_path = toWinPath(dir_path)
        else:
            dir_path = toUnixPath(dir_path)

        def func(transport, dir_path):
            sftp = paramiko.SFTPClient.from_transport(transport)
            dir_stat = None
            try:
                dir_stat = sftp.stat(dir_path).st_mode
            except IOError as e:
                if e.errno == errno.ENOENT:
                    sftp.mkdir(dir_path)
                else:
                    raise e
            else:
                if not stat.S_ISDIR(dir_stat):
                    raise GuestOSException(vm_address, -1, 'Failed to overwrite file/folder. Please make sure the file/folder property is the same.')

        self.runGuestSFTP(func, auth_info, dir_path)

    def downloadFile(self, vm_id=None, dst_path=None, src_path=None):
        options = 'noserverino,iocharset=utf8,password=%s,uid=%d,gid=%d,ro' % \
            (self.auth_info['password'],
            int(os.environ['PKG_UID']),
            int(os.environ['PKG_GID']))

        username = self.auth_info['username'].split('\\')
        if len(username) == 2:
            options += ',domain=%s,username=%s' % (username[0], username[1])
        else:
            options += ',domain=WORKGROUP,username=%s' % (username[0])

        dst_path = dst_path.encode('utf-8')
        host_dir = ('//%s/%s' % (self.auth_info['hostname'], ntpath.dirname(src_path.replace(':', '$', 1)))).encode('utf-8')
        temp_dir = os.path.join(VM_SYNOABK_TEMP, randomString()).encode('utf-8')
        temp_path = os.path.join(temp_dir, ntpath.basename(src_path)).encode('utf-8')

        if 0 <= src_path.find(','):
            raise InvalidArgument('Invalid character "," in path [%s]' % src_path)

        try:
            os.mkdir(temp_dir)

            ret = subprocess.Popen(['/bin/mount.cifs', '-o', options+',vers=syno', host_dir, temp_dir], stderr=subprocess.PIPE).communicate()
            if ret[1]:
                ret = subprocess.Popen(['/bin/mount.cifs', '-o', options, host_dir, temp_dir], stderr=subprocess.PIPE).communicate()
                if ret[1]:
                    raise SambaMountFail(ret[1])

            shutil.copyfile(temp_path, dst_path)
        finally:
            try:
                subprocess.Popen(['/bin/umount', '-l', temp_dir], stderr=subprocess.PIPE).communicate()
            except:
                pass
            try:
                os.rmdir(temp_dir)
            except:
                pass

    def getMigrateCapability(self, host_id=None, host_obj=None):
        return {
            'vmotion': False,
            'vcenter': True
        }

    def getHostInterface(self, host_id=None):
        pass

    def runShell(self, command, timeout):
        return self._run_cmd(command, False)

    def runTaskGuestSFTP(self, task, f, auth_info, *args, **kwargs):
        try:
            return self.runGuestSFTP(f, auth_info, *args, **kwargs)
        except Exception as e:
            task.info.state = 'error'
            task.info.error = e
            raise e

    def runGuestSFTP(self, f, auth_info, *args, **kwargs):
        address, port, username, password = self._get_vm_connection_info(auth_info)

        try:
            transport = paramiko.Transport((address, port))
            transport.default_window_size = paramiko.common.MAX_WINDOW_SIZE
            transport.packetizer.REKEY_BYTES = pow(2, 40)
            transport.packetizer.REKEY_PACKETS = pow(2, 40)

            transport.connect(
                username = username,
                password = password)
            ret = f(transport, *args, **kwargs)
        except (IOError, socket.gaierror) as e:
            raise GuestOSException(address, e.errno, e.strerror)
        except SSHException as e:
            if 'Channel closed.' == e.message:
                raise GuestOSException(address, -1, 'Executing commands on a remote machine over SSH is not available')
            raise GuestOSException(address, -1, str(e))
        finally:
            try:
                transport.close()
            except:
                traceback.print_exc(file=sys.stderr)
                pass

        return ret

    def runGuestShell(self, auth_info, command, timeout=180, chcp=False):
        address, port, username, password = self._get_vm_connection_info(auth_info)

        try:
            ssh = paramiko.SSHClient()
            ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            ssh.connect(
                address,
                port = port,
                username = username,
                password = password)

            if chcp:
                ssh.exec_command("chcp 65001")
            stdin, stdout, stderr = ssh.exec_command(command)

            if timeout > 0:
                elapsed = 0
                while not stdout.channel.exit_status_ready():
                    if elapsed >= timeout:
                       raise TimeoutException(msg='execute [%s] timeout' % command)
                    time.sleep(1)
                    elapsed += 1

            status = stdout.channel.recv_exit_status()
            stdout = stdout.read().strip('\n')
            stderr = stderr.read().strip('\n')

            if status != 0:
                raise GuestOSException(address, status, 'Failed to execute [%s]' % command)
        except (IOError, socket.gaierror) as e:
            raise GuestOSException(address, e.errno, e.strerror)
        except SSHException as e:
            if 'Channel closed.' == e.message:
                raise GuestOSException(address, -1, 'Executing commands on a remote machine over SSH is not available')
            raise GuestOSException(address, -1, str(e))
        finally:
            try:
                ssh.close()
            except:
                traceback.print_exc(file=sys.stderr)
                pass

        return stdout

    ####  util functions ###
    def _get_vm_connection_info(self, auth_info):
        return (
            auth_info['address'] if auth_info['address'] else self._get_vm_address(auth_info['vm_id']),
            auth_info['port'] if auth_info['port'] else 22,
            auth_info['username'],
            auth_info['password']
        )
    def _get_vm_address(self, vm_id):
        address = ''
        network = self._run_cmd(
            'Get-VM -Id %s '
            '| Select -ExpandProperty NetworkAdapters '
            '| Select IPAddresses' % vm_id)
        if 'IPAddresses' in network and 0 < len(network['IPAddresses']):
            address = network['IPAddresses'][0]
        else:
            raise GuestOSException('', -1, 'Can not detect the address of guest OS. Please set it manually.')
        return address

    def _parse_datastore_data(self, datastore_id):
        device_id = ntpath.splitdrive(datastore_id)[0]
        datastore = self._run_cmd('Get-WMIObject -Query "Select * From win32_logicaldisk Where DeviceID=\'%s\'"' % device_id)

        if datastore == None:
            raise Exception('Cannot find datastore [%s]' % datastore_id)

        obj = {}
        obj['name'] = device_id
        obj['accessible'] = True
        obj['capacity'] = datastore['Size']
        obj['free_space'] = datastore['FreeSpace']
        obj['datastore_id'] = obj['id'] = datastore_id
        return obj

    def _parse_snapshot_data(self, vm_id=None, snap_id=None):
        if snap_id:
            cmd = 'Get-VMSnapshot -Id ' + snap_id
        else:
            cmd = 'Get-VM -Id ' + vm_id + '| Get-VMSnapshot'
        cmd += '| Select Id,Name,SnapshotType' \
            + ',@{N="CreateTime";E={Get-Date -Date $_.CreationTime -UFormat %s}}' \
            + ',@{N="VirtualDisk";E={,@($_.HardDrives|Select Path,DiskNumber)}}' \
            + ',@{N="ConfigFiles";E={,@(Get-ChildItem -LiteralPath $_.Path -Filter ($_.Id.Guid + ".*") -File -Recurse|Select FullName)}}'

        snapshots = []
        for snapshot in toList(self._run_cmd(cmd)):
            obj = {}
            obj['snap_id'] = snapshot['Id']
            obj['desc'] = SNAPSHOT_DESC
            obj['name'] = snapshot['Name']
            obj['type'] = snapshot['SnapshotType']
            obj['cbt_enable'] = True
            obj['create_time'] = snapshot['CreateTime']
            obj['virtual_disk'] = self._parse_virtual_disk(snapshot['VirtualDisk'])
            obj['config_files'] = self._parse_config_files(snapshot['ConfigFiles'])
            snapshots.append(obj)

        return snapshots[0] if snap_id else snapshots

    def _parse_vm_data(self, vm_id=None, get_detail=False):
        cmd = 'Get-VM'
        if vm_id:
            cmd += ' -Id ' + vm_id
        cmd += '| Select Id,Name,MemoryStartup,ProcessorCount,State,Status,PrimaryOperationalStatus,ComputerName,Version'
        if get_detail:
            cmd += ',@{N="NetworkAdapters";E={,@($_.NetworkAdapters|Select SwitchId,SwitchName)}},Generation' \
                +  ',@{N="VirtualDisk";E={,@($_.HardDrives|Select Path,DiskNumber)}}' \
                +  ',@{N="ConfigFiles";E={,@(Get-ChildItem -LiteralPath $_.Path -Filter ($_.Id.Guid + ".*") -File -Recurse|Select FullName)}}' \
                +  ',@{N="UsedSpace";E={($_,($_|Get-VMSnapshot)|Get-VMHardDiskDrive|Get-VHD|Measure-Object FileSize -Sum).Sum}}' \
                +  ',@{N="OsName";E={\n' \
                +      '$vm = Get-WmiObject -Namespace root\\virtualization\\v2 -Class Msvm_ComputerSystem | Where Name -eq $_.Id.Guid\n' \
                +      '$vm.GetRelated("Msvm_KvpExchangeComponent").GuestIntrinsicExchangeItems | % {\n' \
                +          '$node = ([XML]$_).SelectSingleNode("/INSTANCE/PROPERTY[@NAME=\'Name\']")\n' \
                +          'if ($node.value -eq "OSName") {\n' \
                +              '$name = $node.SelectSingleNode("/INSTANCE/PROPERTY[@NAME=\'Data\']").value\n' \
                +          '}\n' \
                +          'if ($node.value -eq "ProcessorArchitecture") {\n' \
                +              '$arch = $node.SelectSingleNode("/INSTANCE/PROPERTY[@NAME=\'Data\']").value\n' \
                +          '}\n' \
                +      '}\n' \
                +      'if ($name -and $arch) { $name + " (" + @("64", "32")[$arch -eq 0] + "-bit)" }\n' \
                +  '}}'

        vms = []
        for vm in toList(self._run_cmd(cmd, True, False)):
            try:
                obj = {}
                obj['type'] = 'vim.VirtualMachine'
                obj['name'] = vm['Name']
                obj['vm_id'] = obj['id'] = vm['Id']
                obj['os_name'] = vm['OsName'] if ('OsName' in vm and vm['OsName']) else 'Other'
                obj['is_template'] = False
                obj['vmtool_running'] = True
                obj['instance_uuid'] = vm['Id']
                obj['ram_mb'] = vm['MemoryStartup']/1024/1024
                obj['num_cpu'] = vm['ProcessorCount']
                obj['powered_on'] = vm['State'] == 2 # 2 = running, 3 = off
                obj['status'] = vm['Status']
                obj['status_ok'] = vm['PrimaryOperationalStatus'] == 2
                obj['hw_version'] = vm['Version']

                if get_detail:
                    obj['folder_path'] = ''
                    obj['hypervisor_name'] = ''
                    obj['storage_size'] = vm['UsedSpace']
                    obj['virtual_disk'] = self._parse_virtual_disk(vm['VirtualDisk'], vm_id is None)
                    obj['config_files'] = self._parse_config_files(vm['ConfigFiles'])

                    obj['hypervisor'] = {
                        'name': vm['ComputerName'],
                        'host_id': '',
                        'standalone': True
                    }
                    obj['resourcepool'] = {
                        'is_vapp': False,
                        'name': vm['ComputerName'],
                        'resourcepool_id': vm['ComputerName']
                    }
                    if len(obj['config_files']) > 0:
                        folder_id = ntpath.dirname(ntpath.dirname(obj['config_files'][0]))
                        obj['folder'] = {
                            'name': ntpath.basename(folder_id),
                            'folder_id': toWinPath(folder_id),
                            'datacenter_id': ''
                        }
                    else:
                        obj['folder'] = {
                            'name': '',
                            'folder_id': '',
                            'datacenter_id': ''
                        }

                    obj['datastore'] = []
                    for disk in obj['virtual_disk']:
                        datastore_id = ntpath.dirname(disk['filename'])
                        obj['datastore'].append({
                            'name': ntpath.basename(datastore_id),
                            'datastore_id': toWinPath(datastore_id)
                        })
                    obj['network'] = []
                    for network in vm['NetworkAdapters']['value']:
                        obj['network'].append({
                            'name': network['SwitchName'],
                            'network_id': network['SwitchId']
                        })
                    obj['is_bios'] = vm['Generation'] == 1
                    obj['cbt_enabled'] = True

                vms.append(obj)
            except Exception:
                if vm_id:
                    raise
                else:
                    traceback.print_exc(file=sys.stderr)

        return vms[0] if vm_id else vms

    def _parse_config_files(self, data):
        if not data or 'value' not in data:
            return []

        files = []
        for file in data['value']:
            file = file['FullName'].lower()
            if file.endswith('vmcx') or file.endswith('vmrs') or file.endswith('vmgs'):
                files.append(toUnixPath(file))
        return files

    def _parse_virtual_disk(self, data, no_throw = False):
        if not data or 'value' not in data:
            return []

        disks = []
        for disk in data['value']:
            obj = {}
            if not disk['Path']:
                continue
            try:
                if '\\' in disk['Path']:
                    vhd = self._run_cmd('Get-VHD -Path "' + escapePath(disk['Path']) + '" | Select Path,Size,@{N="Root";E={\n' \
                        +     '$path = $_.Path\n' \
                        +     'while ($path) {\n' \
                        +         '$disk = Get-VHD -Path $path\n' \
                        +         '$path = $disk.ParentPath\n' \
                        +     '}\n' \
                        +     '$disk|Select Path,VhdType,DiskIdentifier\n' \
                        + '}}')

                    vhd['Path'] = vhd['Path'].lower()
                    vhd['Root']['Path'] = vhd['Root']['Path'].lower()
                    obj['key'] = vhd['Root']['DiskIdentifier']
                    obj['name'] = ntpath.basename(vhd['Root']['Path'])
                    obj['eager_scrub'] = True
                    obj['capacity'] = vhd['Size']
                    obj['filename'] = toUnixPath(vhd['Path'])
                    obj['filename_root'] = toUnixPath(vhd['Root']['Path'])
                    obj['thin_provision'] = vhd['Root']['VhdType'] == 3 # dynamic = 3, fixed = 2
                else:
                    vhd = self._run_cmd('Get-WmiObject Win32_DiskDrive | Where Index -eq ' + str(disk['DiskNumber']) + ' | Select Size')
                    obj['key'] = disk['Path'].lower()
                    obj['name'] = ''
                    obj['eager_scrub'] = True
                    obj['capacity'] = vhd['Size']
                    obj['filename'] = ''
                    obj['filename_root'] = disk['Path']
                    obj['thin_provision'] = False
                    obj['disk_mode'] = 'passthrough'
                disks.append(obj)
            except:
                if not no_throw:
                    raise
                traceback.print_exc(file=sys.stderr)
                continue

        return disks

    def _run_cmd(self, cmd, to_json=True, err_stop=True, long_cmd=False):
        if to_json:
            cmd += '| ConvertTo-Json -Depth 10'

        cmd = '$ErrorActionPreference = "%s"\n' % ('Stop' if err_stop else 'Continue') \
            + 'Try {\n' \
            +     cmd + '\n' \
            + '} Catch {\n' \
            +     '%s\n' % ('' if long_cmd else 'Write-Host ($_ | ConvertTo-Json)') \
            +     'throw\n' \
            + '}'

        if long_cmd:
            sha1 = hashlib.sha1()
            sha1.update(cmd)
            basename = sha1.hexdigest() + '.ps1'
            src_path = os.path.join(VM_SYNOABK_TEMP, basename).encode('utf-8')
            dst_root = self._run_cmd('echo $env:LOCALAPPDATA/ActiveBackup', False)
            dst_path = os.path.join(dst_root, basename).encode('utf-8')

            with codecs.open(src_path, 'w', 'utf-8') as file:
                file.write(u'\ufeff') # utf8 with BOM
                file.write(cmd)

            try:
                self.createFolder('', dst_root)
            except:
                pass

            try:
                self.uploadFile('', '', src_path, dst_path)
                std_out = self._run_cmd(dst_path, False, False, False)
                status_code = 0
            finally:
                try:
                    os.remove(src_path)
                except:
                    pass
                try:
                    self.removeFolder('', dst_path)
                except:
                    pass
        else:
            enc_ps = b64encode(cmd.encode('utf_16_le')).decode('ascii')
            cmd_id = self.protocol.run_command(self.shell_id, 'powershell -encodedcommand %s' % enc_ps)
            std_out, std_err, status_code = self.protocol.get_command_output(self.shell_id, cmd_id)
            self.protocol.cleanup_command(self.shell_id, cmd_id)

        if status_code:
            raise PowerShellException(std_out)
        if to_json and std_out:
            return json.loads(std_out)
        if std_out:
            return std_out.strip().decode('utf-8-sig', 'ignore')
        return None

    def _get_task(self, task_id):
        return self.task_pool[task_id]

    def _put_task(self, task):
        self.task_pool[task.info.key] = task
        return task.info.key

    def _del_task(self, task_id):
        del self.task_pool[task_id]
